/*
 * puzzle1.c
 *
 *
 * Part of TREE-PUZZLE 5.0 (June 2000)
 *
 * (c) 1999-2000 by Heiko A. Schmidt, Korbinian Strimmer,
 *                  M. Vingron, and Arndt von Haeseler
 * (c) 1995-1999 by Korbinian Strimmer and Arndt von Haeseler
 *
 * All parts of the source except where indicated are distributed under
 * the GNU public licence.  See http://www.opensource.org for details.
 */


#define EXTERN
#define FILENAMELEN 256

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <limits.h>
#include <float.h>

unsigned char readquartet(int a, int b, int c, int d, 
                          unsigned char *quartetinfo);
void writequartet(int a, int b, int c, int d, unsigned char info,
                  unsigned char *quartetinfo);
void readquartetlh(int a, int b, int c, int d, 
                   double *d1, double *d2, double *d3, 
                   double *lhinfo);
void writequartetlh(int a, int b, int c, int d, 
                    double d1, double d2, double d3,
                    double *lhinfo);
unsigned char weightquart(double d1, double d2, double d3);




char topostr[8][4] = {"000", "001", "010", "011",
                      "100", "101", "110", "111"};

typedef unsigned long int uli;

typedef struct {
  char           *fname;
  char            format[2];
  FILE           *fstream;
  int             nspec;
  int            *translat;
  char          **names;
  unsigned char  *quartetinfo;
  double         *lhinfo;
} qblocktype;

/*********************************************************************
*  read/write array with all quartets                                *
*********************************************************************/


void sort4ints(int *a, int *b, int *c, int *d)
{
  int t;
  if (*a > *b) { t=*a; *a=*b; *b=t; }
  if (*a > *c) { t=*a; *a=*c; *c=t; }
  if (*a > *d) { t=*a; *a=*d; *d=t; }
  
  if (*b > *c) { t=*b; *b=*c; *c=t; }
  if (*b > *d) { t=*b; *b=*d; *d=t; }
  
  if (*c > *d) { t=*c; *c=*d; *d=t; }
}

/******************************************************************************/

void sort3doubles(double num[3], int order[3])
{
        if (num[0] > num[1]) {
                if(num[2] > num[0]) {
                        order[0] = 2;
                        order[1] = 0;
                        order[2] = 1;
                } else if (num[2] < num[1]) {
                        order[0] = 0;
                        order[1] = 1;
                        order[2] = 2;
                } else {
                        order[0] = 0;
                        order[1] = 2;
                        order[2] = 1;
                }
        } else {
                if(num[2] > num[1]) {
                        order[0] = 2;
                        order[1] = 1;
                        order[2] = 0;
                } else if (num[2] < num[0]) {
                        order[0] = 1;
                        order[1] = 0;
                        order[2] = 2;
                } else {
                        order[0] = 1;
                        order[1] = 2;
                        order[2] = 0;
                }
        }
} 


/******************************************************************************/

void num2quart(uli qnum, int *a, int *b, int *c, int *d)
{
        double temp;
        uli aa, bb, cc, dd;
        uli lowval=0, highval=0;

        aa=0; bb=1; cc=2; dd=3;

        temp = (double)(24 * qnum);
        temp = sqrt(temp);
        temp = sqrt(temp);
        /* temp = pow(temp, (double)(1/4)); */
        dd = (uli) floor(temp) + 1;
        if (dd < 3) dd = 3;
        lowval =  (uli) dd*(dd-1)*(dd-2)*(dd-3)/24;
        highval = (uli) (dd+1)*dd*(dd-1)*(dd-2)/24;
        if (lowval >= qnum)
            while ((lowval > qnum)) {
                dd -= 1; lowval = (uli) dd*(dd-1)*(dd-2)*(dd-3)/24;
            }
        else {
            while (highval <= qnum) {
                dd += 1; highval = (uli) (dd+1)*dd*(dd-1)*(dd-2)/24;
            }
            lowval = (uli) dd*(dd-1)*(dd-2)*(dd-3)/24;
        }
        qnum -= lowval;
        if (qnum > 0) {
            temp = (double)(6 * qnum);
            temp = pow(temp, (double)(1/3));
            cc = (uli) floor(temp);
            if (cc < 2) cc= 2;
            lowval =  (uli) cc*(cc-1)*(cc-2)/6;
            highval = (uli) (cc+1)*cc*(cc-1)/6;
            if (lowval >= qnum)
                while ((lowval > qnum)) {
                   cc -= 1; lowval = (uli) cc*(cc-1)*(cc-2)/6;
                }
            else {
                while (highval <= qnum) {
                   cc += 1; highval = (uli) (cc+1)*cc*(cc-1)/6;
                }
                lowval = (uli) cc*(cc-1)*(cc-2)/6;
            }
            qnum -= lowval;
            if (qnum > 0) {
                temp = (double)(2 * qnum);
                temp = sqrt(temp);
                bb = (uli) floor(temp);
                if (bb < 1) bb= 1;
                lowval =  (uli) bb*(bb-1)/2;
                highval = (uli) (bb+1)*bb/2;
                if (lowval >= qnum)
                    while ((lowval > qnum)) {
                        bb -= 1; lowval = (uli) bb*(bb-1)/2;
                    }
                else {
                    while (highval <= qnum) {
                       bb += 1; highval = (uli) (bb+1)*bb/2;
                    }
                    lowval = (uli) bb*(bb-1)/2;
                }
                qnum -= lowval;
                if (qnum > 0) {
                   aa = (uli) qnum;
                   if (aa < 0) aa= 0;
                }
            }
        }
        *d = (int)dd;
        *c = (int)cc;
        *b = (int)bb;
        *a = (int)aa;
}  /* num2quart */

/******************/

uli numquarts(int maxspc)
{
   uli tmp;
   int a, b, c, d;

   if (maxspc < 4)
     return (uli)0;
   else {
      maxspc--;
      a = maxspc-3;
      b = maxspc-2;
      c = maxspc-1;
      d = maxspc;

      tmp = (uli) 1 + a +
            (uli) b * (b-1) / 2 +
            (uli) c * (c-1) * (c-2) / 6 +
            (uli) d * (d-1) * (d-2) * (d-3) / 24;
      return (tmp);
   }
}  /* numquarts */

/******************/

uli quartnum2bytes(int maxquart)
{
  /* compute number of bytes */
  if (maxquart % 2 == 0) { /* even number */
     return((maxquart)/2);
  } else { /* odd number */
     return((maxquart + 1)/2);
  }

}

/******************/

uli specnum2bytes(int maxspc)
{
  return (quartnum2bytes(numquarts(maxspc)));
}

/******************/

uli quart2num (int a, int b, int c, int d)
{
      uli tmp;
      if ((a>b) || (b>c) || (c>d)) {
         fprintf(stderr, "Warning PP5: not (%d<%d<%d<%d) !!!\n",a,b,c,d);
         sort4ints(&a, &b, &c, &d);
      }
      tmp = (uli) a +
            (uli) b * (b-1) / 2 +
            (uli) c * (c-1) * (c-2) / 6 +
            (uli) d * (d-1) * (d-2) * (d-3) / 24;
      return (tmp);
}  /* quart2num */

/******************/

void printallquarts(	FILE       *fp, 
			qblocktype  qb, 
			int         idx, 
			int         brackets, 
			int         topol, 
			int         bits, 
			int         names,
			int         formatflag);


void printverboseallquarts(FILE *fp, qblocktype qb)
{
  printallquarts(	fp, 
			qb,
			1,  /* idx */ 
			1,  /* brackets */ 
			1,  /* topol */ 
			1,  /* bits */ 
			1,  /* names */
			0   /* 0-old, 1-BB, 2-BA, 3-LB, 4-LA */);
}


void printasciiallquarts(FILE *fp, qblocktype qb)
{
  printallquarts(	fp, 
			qb,
			0,  /* idx */ 
			1,  /* brackets */ 
			0,  /* topol */ 
			1,  /* bits */ 
			0,  /* names */
			2   /* 0-old, 1-BB, 2-BA, 3-LB, 4-LA */);
}


/******************/

/*  flag=0	old allquart binary  */
/*  flag=1	allquart binary  */
/*  flag=2	allquart ACSII   */
/*  flag=3	quartlh  binary  */
/*  flag=4	quartlh  ASCII   */

void writetpqfheader(int            nspec,
                     char         **names,
                     FILE          *ofp,
                     int            flag)
{ int            currspec;

  if (flag == 0) {
     unsigned long  nquart;
     unsigned long  blocklen;

     nquart = numquarts(nspec);
     /* compute number of bytes */
     if (nquart % 2 == 0) { /* even number */
        blocklen = (nquart)/2;
     } else { /* odd number */
        blocklen = (nquart + 1)/2;
     }
     /* FPRINTF(STDOUTFILE "Writing quartet file: %s\n", filename); */
     fprintf(ofp, "TREE-PUZZLE\n5.0\n\n");
     fprintf(ofp, "species: %d\n",       nspec);
     fprintf(ofp, "quartets: %lu\n",     nquart);
     fprintf(ofp, "bytes: %lu\n\n",      blocklen);


     /* fwrite(&(quartetinfo[0]), sizeof(char), blocklen, ofp); */
  }

  if (flag == 1) fprintf(ofp, "##TPQF-BB (CHECKQUART)\n%d\n", nspec);
  if (flag == 2) fprintf(ofp, "##TPQF-BA (CHECKQUART)\n%d\n", nspec);
  if (flag == 3) fprintf(ofp, "##TPQF-LB (CHECKQUART)\n%d\n", nspec);
  if (flag == 4) fprintf(ofp, "##TPQF-LA (CHECKQUART)\n%d\n", nspec);

  for (currspec=0; currspec<nspec; currspec++) {
     fprintf(ofp, "%s\n", names[currspec]);
  } /* for each species */
  fprintf(ofp, "\n");

} /* writetpqfheader */



/******************/

void printallquarts(	FILE       *fp, 
			qblocktype  qb, 
			int         idx, 
			int         brackets, 
			int         topol, 
			int         bits, 
			int         names,
			int         formatflag)
{
  int a, b, c, d;
  uli qnum, qbytes;
  unsigned topos;

  qnum = numquarts(qb.nspec);
  qbytes = quartnum2bytes(qnum);

  writetpqfheader(qb.nspec, qb.names, fp, formatflag);

  if (qb.quartetinfo == NULL)
    fprintf(fp, "NO QUARTETBLOCK READ!\n");
  else
    for (d = 3; d < qb.nspec; d++)
      for (c = 2; c < d; c++)
        for (b = 1; b < c; b++)
          for (a = 0; a < b; a++) {
	    topos = readquartet(a, b, c, d, qb.quartetinfo);
	    if (idx)
		fprintf(fp, "%lu\t", quart2num(a, b, c, d));
	    if (brackets)
		fprintf(fp, "(%d,%d,%d,%d)\t", a, b, c, d);
	    if (topol)
		fprintf(fp, "%d\t", topos);
	    if (bits)
		fprintf(fp, "%s\t", topostr[topos]);
	    if (names)
		fprintf(fp, "%s,%s,%s,%s", qb.names[a], qb.names[b],
                                           qb.names[c], qb.names[d]);
  		fprintf(fp, "\n"); fflush(stdout);
          }
}

/************************/

void printbinlh(qblocktype *qb)
{
  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "w");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s to write!\n\n", qb->fname);
      exit (1);
    }
  }

  writetpqfheader(qb->nspec, qb->names, qb->fstream, 3);

  if (qb->lhinfo == NULL)
    fprintf(stderr, "NO QUARTETBLOCK READ!\n");
  else
    fwrite(qb->lhinfo, sizeof(double), 3 * numquarts(qb->nspec), qb->fstream);

  fclose(qb->fstream);
  qb->fstream = NULL;
}

/************************/

void printbinallquarts(qblocktype *qb)
{
  uli qnum, qbytes;
  int            currspec;

  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "w");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s to write!\n\n", qb->fname);
      exit (1);
    }
  }

  fprintf(qb->fstream, "TREE-PUZZLE\n");
  fprintf(qb->fstream, "5.0\n");

  qnum = numquarts(qb->nspec);
  qbytes = quartnum2bytes(qnum);
  fprintf(qb->fstream, "\n");
  fprintf(qb->fstream, "species: %d\n", qb->nspec);
  fprintf(qb->fstream, "quartets: %lu\n", qnum);
  fprintf(qb->fstream, "bytes: %lu\n", qbytes);
  fprintf(qb->fstream, "\n");

  for (currspec=0; currspec < (qb->nspec); currspec++) 
     fprintf(qb->fstream, "%s\n", (qb->names)[currspec]);
  fprintf(qb->fstream, "\n");

  if (qb->quartetinfo == NULL)
    fprintf(stderr, "NO QUARTETBLOCK READ!\n");
  else
    fwrite(qb->quartetinfo, sizeof(char), qbytes, qb->fstream);

  fclose(qb->fstream);
  qb->fstream = NULL;
}

/************************/

void newreadheader(qblocktype *qb, int closeflag)
{ 
  int           currspec;
  char          dummyprog[128];
  char		tmpchr;
  int           strp;
  int           flag;


  if (qb->fstream != NULL) {
    fclose(qb->fstream);
    qb->fstream = NULL;
  }

  qb->fstream = fopen(qb->fname, "r");
  if (qb->fstream == NULL) {
    fprintf( stderr, "\nCan't open file %s\n\n", qb->fname);
    exit (1);
  }

  fscanf(qb->fstream, "##TPQF-%c%c\n", &(qb->format[0]), &(qb->format[1]));
  flag = 0;
  strp = 0;
  tmpchr = fgetc(qb->fstream);
  while (tmpchr != '\n') {
     if (!flag && (tmpchr == '('))
          flag = 1;
     else if (tmpchr == ')')
               flag = 0;
          else if (flag)
                  dummyprog[strp++] = tmpchr;
     tmpchr = fgetc(qb->fstream);
  }
  dummyprog[strp] = '\0';
  fscanf(qb->fstream, "%d\n",   &(qb->nspec));

  fprintf(stdout, "Reading quartet file: %s\n", qb->fname);
  fprintf(stdout, "   type: %c%c\n", qb->format[0], qb->format[1]);
  fprintf(stdout, "   generated by: %s\n", dummyprog);

  fprintf(stdout, "   %d species\n", qb->nspec);

  qb->names       = malloc(qb->nspec * sizeof(char *));

  for (currspec=0; currspec<qb->nspec; currspec++) {

     (qb->names)[currspec] = (char *)malloc(11 * sizeof(char));
     fscanf(qb->fstream, "%s\n", (qb->names)[currspec]);
     fprintf(stdout, "   %3d. %s\n", currspec+1, (qb->names)[currspec]);

  } /* for each species */
  fprintf(stdout, "\n");

  fscanf(qb->fstream, "\n");

  if (closeflag) {
     fclose(qb->fstream);
     qb->fstream = NULL;
  }

} /* readallquart */

/************************/

void newreadallquarts(qblocktype    *qb)
{ unsigned long nquart;
  unsigned long blocklen;
  int           a,b,c,d;
#ifdef VERBOSE
  double        d1,d2,d3;
#endif
  float         l1,l2,l3;
  unsigned long n;

  newreadheader(qb, 0);

  nquart = numquarts(qb->nspec);

  if ((qb->format[0] == 'B') && (qb->format[1] == 'B')) {   /* topology bits in binary format */
      fprintf(stderr, "%c%c\n",qb->format[0],qb->format[1]);
      qb->quartetinfo = malloc(specnum2bytes(qb->nspec) * sizeof(unsigned char));
      fread(&(qb->quartetinfo[0]), sizeof(unsigned char), blocklen, qb->fstream);
  }

  if ((qb->format[0] == 'L') && (qb->format[1] == 'B')) {   /* log-LH in binary format */
      fprintf(stderr, "%c%c\n",qb->format[0],qb->format[1]);
      qb->lhinfo = malloc(3 * numquarts(qb->nspec) * sizeof(double));
      fread(&(qb->lhinfo[0]), sizeof(double), 3*numquarts(qb->nspec), qb->fstream);
#ifdef VERBOSE
      for (n=0; n<nquart; n++){
         num2quart(n, &a, &b, &c, &d);
         readquartetlh(a, b, c, d, &d1, &d2, &d3, qb->lhinfo);
         printf("(%d,%d,%d,%d)\t%f\t%f\t%f\n",a,b,c,d, d1, d2, d3);
      }
#endif
  }

  if ((qb->format[0] == 'L') && (qb->format[1] == 'A')) {   /* log-LH in ASCII format */
      fprintf(stderr, "%c%c\n",qb->format[0],qb->format[1]);
      qb->lhinfo = malloc(3 * numquarts(qb->nspec) * sizeof(double));
      /* fread(&(qb->hlinfo[0]), sizeof(unsigned char), blocklen, qb->fstream); */
      for (n=0; n<nquart; n++){
         fscanf(qb->fstream, "(%d,%d,%d,%d)\t%f\t%f\t%f\n",&a,&b,&c,&d, &l1, &l2, &l3);
#ifdef VERBOSE
         printf("(%d,%d,%d,%d)\t%f\t%f\t%f\n",a,b,c,d, l1, l2, l3);
#endif
         writequartetlh(a, b, c, d, (double)l1, (double)l2, (double)l3, qb->lhinfo);
      }
  }

  fclose(qb->fstream);
  qb->fstream = NULL;

} /* readallquart */

/************************/

void freshreadheader(qblocktype    *qb)
{ unsigned long nquart;
  unsigned long blocklen;
  int           currspec;
  char          dummyversion[128];
  char          dummyprog[128];


  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "r");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s\n\n", qb->fname);
      exit (1);
    }
  }
  fscanf(qb->fstream, "%s\n", dummyprog);
  fscanf(qb->fstream, "%s\n\n", dummyversion);
  
  fscanf(qb->fstream, "species: %d\n",   &(qb->nspec));
  fscanf(qb->fstream, "quartets: %lu\n", &nquart);
  fscanf(qb->fstream, "bytes: %lu\n\n",  &blocklen);

  fprintf(stdout, "Reading quartet file: %s\n", qb->fname);
  fprintf(stdout, "   generated by: %s %s\n", dummyprog, dummyversion);

  fprintf(stdout, "   %d species, %lu quartets, %lu bytes\n", 
                   qb->nspec, nquart, blocklen);

  qb->names       = malloc(qb->nspec * sizeof(char *));

  for (currspec=0; currspec<qb->nspec; currspec++) {

     (qb->names)[currspec] = (char *)malloc(11 * sizeof(char));
     fscanf(qb->fstream, "%s\n", (qb->names)[currspec]);
     fprintf(stdout, "   %3d. %s\n", currspec+1, (qb->names)[currspec]);

  } /* for each species */
  fprintf(stdout, "\n");

/*
  qb->quartetinfo = malloc(blocklen * sizeof(unsigned char));
  fread(&(qb->quartetinfo[0]), sizeof(unsigned char), blocklen, qb->fstream); 
*/
  fclose(qb->fstream);
  qb->fstream = NULL;

} /* readallquart */

/************************/

void freshreadallquarts(qblocktype    *qb)
{ unsigned long nquart;
  unsigned long blocklen;
  int           currspec;
  char          dummyversion[128];
  char          dummyprog[128];


  if (qb->fstream == NULL) {
    qb->fstream = fopen(qb->fname, "r");
    if (qb->fstream == NULL) {
      fprintf( stderr, "\nCan't open file %s\n\n", qb->fname);
      exit (1);
    }
  }
  fscanf(qb->fstream, "%s\n", dummyprog);
  fscanf(qb->fstream, "%s\n\n", dummyversion);
  
  fscanf(qb->fstream, "species: %d\n",   &(qb->nspec));
  fscanf(qb->fstream, "quartets: %lu\n", &nquart);
  fscanf(qb->fstream, "bytes: %lu\n\n",  &blocklen);

  fprintf(stdout, "Reading quartet file: %s\n", qb->fname);
  fprintf(stdout, "   generated by: %s %s\n", dummyprog, dummyversion);

  fprintf(stdout, "   %d species, %lu quartets, %lu bytes\n", 
                   qb->nspec, nquart, blocklen);

  qb->quartetinfo = malloc(blocklen * sizeof(unsigned char));
  qb->names       = malloc(qb->nspec * sizeof(char *));

  for (currspec=0; currspec<qb->nspec; currspec++) {

     (qb->names)[currspec] = (char *)malloc(11 * sizeof(char));
     fscanf(qb->fstream, "%s\n", (qb->names)[currspec]);
     fprintf(stdout, "   %3d. %s\n", currspec+1, (qb->names)[currspec]);

  } /* for each species */
  fprintf(stdout, "\n");

  fread(&(qb->quartetinfo[0]), sizeof(unsigned char), blocklen, qb->fstream);
  fclose(qb->fstream);
  qb->fstream = NULL;

} /* freshreadallquart */

/************************/


void readquartetlh(int a, int b, int c, int d, 
                   double *d1, double *d2, double *d3, double *lhinfo)
{
        uli qnum;

	sort4ints(&a,&b,&c,&d);
        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        qnum *= 3;
        *d1 = lhinfo[qnum];
        *d2 = lhinfo[qnum + 1];
        *d3 = lhinfo[qnum + 2];
} /* readquartetlh */

/************************/

void writequartetlh(int a, int b, int c, int d, 
                    double d1, double d2, double d3,
                    double *lhinfo)
{
        uli qnum;

        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        qnum *= 3;
        lhinfo[qnum]     = d1;
        lhinfo[qnum + 1] = d2;
        lhinfo[qnum + 2] = d3;
} /* writequartetlh */

/************************/

unsigned char readquartet(int a, int b, int c, int d, 
                          unsigned char *quartetinfo)
{
        uli qnum;

	sort4ints(&a,&b,&c,&d);
        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        if (qnum % 2 == 0) { /* even number */
                /* bits 0 to 3 */
                return (quartetinfo[qnum/2] & (unsigned char) 15);
        } else { /* odd number */
                /* bits 4 to 7 */
                return ((quartetinfo[(qnum-1)/2] & (unsigned char) 240)>>4);
        }
} /* readquartet */

/************************/

void writequartet(int a, int b, int c, int d, unsigned char info,
                  unsigned char *quartetinfo)
{
        uli qnum;

        qnum = (uli) a
                        + (uli) b*(b-1)/2
                        + (uli) c*(c-1)*(c-2)/6
                        + (uli) d*(d-1)*(d-2)*(d-3)/24;
        if (qnum % 2 == 0) { /* even number */
                /* bits 0 to 3 */
                quartetinfo[qnum/2] =
                        ((quartetinfo[qnum/2] & (unsigned char) 240) |
                        (info & (unsigned char) 15));
        } else { /* odd number */
                /* bits 4 to 7 */
                quartetinfo[(qnum-1)/2] =
                        ((quartetinfo[(qnum-1)/2] & (unsigned char) 15) |
                        ((info & (unsigned char) 15)<<4));
        }
} /* writequartet */

/******************************************************************************/

void initqblock(qblocktype *qb)
{
  qb->fname       = malloc(FILENAMELEN * sizeof(char));
  qb->format[0]   = '-';
  qb->format[1]   = '-';
  qb->fstream     = NULL;
  qb->nspec       = 0;
  qb->translat    = NULL;
  qb->names       = NULL;
  qb->quartetinfo = NULL;
  qb->lhinfo      = NULL;
}

/************************/

void cleanupqblock(qblocktype *qb)
{
  int n;

  if (qb->fname != NULL) 
      strcpy(qb->fname,"");
  if (qb->fstream != NULL) 
      fclose(qb->fstream);
  qb->fstream = NULL;

  if (qb->names != NULL) {
      for (n=0 ; n < qb->nspec; n++) 
           free((qb->names)[n]);
      free(qb->names);
  }

  qb->nspec   = 0;

  if (qb->translat != NULL)
      free(qb->translat);
  qb->translat = NULL;
  
  if (qb->quartetinfo != NULL)
      free(qb->quartetinfo);
  qb->quartetinfo = NULL;
  
  if (qb->lhinfo != NULL)
      free(qb->lhinfo);
  qb->lhinfo = NULL;
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

void toasciiquart(int argc, char *argv[])
{
  qblocktype      qblock1;
  FILE *ofp;
  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  freshreadallquarts(&qblock1);
  if (argc >= 4) {
    ofp = fopen(argv[3], "w");
    if (ofp == NULL) {
      fprintf( stderr, "\nCan't open file %s to write!\n\n", argv[3]);
      exit (1);
    }
  } else {
      fprintf( stderr, "\nAt least 2 filenames needed!\n\n");
  }
  printasciiallquarts(ofp, qblock1);
  cleanupqblock(&qblock1);
}

/******************************************************************************/

void emptylh(int argc, char *argv[])
{
  qblocktype      qblock1;
  FILE            *ofp;
  int a,b,c,d;

  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  newreadheader(&qblock1, 0);

  ofp = fopen(argv[3], "w");
  if (ofp == NULL) {
    fprintf( stderr, "\nCan't open file %s to write!\n\n", argv[3]);
    exit (1);
  }
  writetpqfheader(qblock1.nspec, qblock1.names, ofp, 4);
  for (d = 3; d < qblock1.nspec; d++)
     for (c = 2; c < d; c++)
        for (b = 1; b < c; b++)
           for (a = 0; a < b; a++)
	      fprintf(ofp, "(%d,%d,%d,%d)\t0.0\t0.0\t0.0\n", a, b, c, d);
  fclose(ofp);
  /* printverboseallquarts(stdout, qblock1); */
  cleanupqblock(&qblock1);
}
/******************************************************************************/

void emptybinlh(int argc, char *argv[])
{
  qblocktype      qblock1;
  FILE            *ofp;
  int a,b,c,d;
  double          dummy[3];

  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  newreadheader(&qblock1, 0);

  ofp = fopen(argv[3], "w");
  if (ofp == NULL) {
    fprintf( stderr, "\nCan't open file %s to write!\n\n", argv[3]);
    exit (1);
  }
  writetpqfheader(qblock1.nspec, qblock1.names, ofp, 3);
  for (d = 3; d < qblock1.nspec; d++)
     for (c = 2; c < d; c++)
        for (b = 1; b < c; b++)
           for (a = 0; a < b; a++)
	      /* fprintf(ofp, "(%d,%d,%d,%d)\t0.0\t0.0\t0.0\n", a, b, c, d); */
              fwrite(&(dummy[0]), sizeof(double), 3, ofp);
  fclose(ofp);
  /* printverboseallquarts(stdout, qblock1); */
  cleanupqblock(&qblock1);
}
/******************************************************************************/

void emptyoldquart(int argc, char *argv[])
{
  qblocktype      qblock1;
  FILE            *ofp;
  int a,b,c,d;
  uli max, n;
  unsigned char dummy[1];

  dummy[0] = 0;
  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  newreadheader(&qblock1, 0);

  ofp = fopen(argv[3], "w");
  if (ofp == NULL) {
    fprintf( stderr, "\nCan't open file %s to write!\n\n", argv[3]);
    exit (1);
  }
  writetpqfheader(qblock1.nspec, qblock1.names, ofp, 0);

  max = specnum2bytes(qblock1.nspec);
  for (n = 0; n < max; n++)
    fwrite(&(dummy[0]), sizeof(unsigned char), 1, ofp);
  fclose(ofp);
  /* printverboseallquarts(stdout, qblock1); */
  cleanupqblock(&qblock1);
}

/******************************************************************************/


void checklh(int argc, char *argv[])
{
  qblocktype      qblock1;
  int a,b,c,d;
  double d1,d2,d3;
  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  newreadallquarts(&qblock1);
  for (d = 3; d < qblock1.nspec; d++) 
    for (c = 2; c < d; c++)
      for (b = 1; b < c; b++)
        for (a = 0; a < b; a++) {
           readquartetlh(a, b, c, d, &d1, &d2, &d3, qblock1.lhinfo);
           printf("(%d,%d,%d,%d)\t%f\t%f\t%f\n",a,b,c,d, d1, d2, d3);
        }

  /* printverboseallquarts(stdout, qblock1); */
  cleanupqblock(&qblock1);
}
/******************************************************************************/

void lh2quart(int argc, char *argv[])
{
  qblocktype      qblock1;
  int a,b,c,d;
  double d1,d2,d3;
  unsigned char tp;

  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  newreadallquarts(&qblock1);

  printf("xxx\n");

  qblock1.quartetinfo = malloc(specnum2bytes(qblock1.nspec));
  strcpy(qblock1.fname, argv[3]);

  for (d = 3; d < qblock1.nspec; d++) {
    fprintf(stderr, "%d\n",d); 

    for (c = 2; c < d; c++)
      for (b = 1; b < c; b++)
        for (a = 0; a < b; a++) {
           readquartetlh(a,b,c,d, &d1, &d2, &d3, qblock1.lhinfo);
           if ((d1==0.0) && (d2==0.0) & (d3==0.0))
               tp = 7;
           else
               tp = weightquart(d1, d2, d3);
           writequartet(a,b,c,d, tp, qblock1.quartetinfo);
           tp = readquartet(a,b,c,d, qblock1.quartetinfo);
           /* printf("(%d,%d,%d,%d)\t%f\t%f\t%f\t%d\n",a,b,c,d, d1, d2, d3, tp); */
        }
  }
  /* printverboseallquarts(stdout, qblock1); */
  printbinallquarts(&qblock1);
  cleanupqblock(&qblock1);
}
/******************************************************************************/

void checkquart(int argc, char *argv[])
{
  qblocktype      qblock1;
  initqblock(&qblock1);
  strcpy(qblock1.fname, argv[2]);
  freshreadallquarts(&qblock1);
  printverboseallquarts(stdout, qblock1);
  cleanupqblock(&qblock1);
}

/******************************************************************************/

void mergequart(int argc, char *argv[])
{
  qblocktype      qblock1;
  qblocktype      allqblock;
  int             currfile,
                  n1,
                  n2;
  int             sumspec = 0;

  printf(" argc=%d\n", argc);
  initqblock(&allqblock);

  for (currfile=2; currfile < argc; currfile++) {
    initqblock(&qblock1);
    strcpy(qblock1.fname, argv[currfile]);
    /* freshreadallquarts(&qblock1); */
    freshreadheader(&qblock1);
    printverboseallquarts(stdout, qblock1);
    
    sumspec += qblock1.nspec;
    cleanupqblock(&qblock1);
  }


  printf("sum=%d\n", sumspec);
  
  allqblock.names = malloc(sumspec * sizeof(char **));
  allqblock.nspec = 0;
  for (currfile=2; currfile < argc; currfile++) {
    initqblock(&qblock1);
    strcpy(qblock1.fname, argv[currfile]);
    freshreadallquarts(&qblock1);
    qblock1.translat = malloc(qblock1.nspec * sizeof(int));
    for (n1=0; n1<qblock1.nspec; n1++) {
      qblock1.translat[n1] = -1;
      for (n2=0; n2<allqblock.nspec; n2++) {
         if (0 == strcmp(qblock1.names[n1], allqblock.names[n2])) {
           fprintf(stdout, "%3d %10s == %3d %10s\n", n1, qblock1.names[n1], n2, allqblock.names[n2]);
           qblock1.translat[n1] = n2;
         }
      }
      if (qblock1.translat[n1] == -1) {
        allqblock.names[allqblock.nspec] = malloc(11 * sizeof(char));
        strcpy(allqblock.names[allqblock.nspec], qblock1.names[n1]);
        qblock1.translat[n1] = allqblock.nspec;
        allqblock.nspec++;
      }
    }
  }

/* printf("aaa1\n"); fflush(stdout); */
  allqblock.quartetinfo = malloc(specnum2bytes(allqblock.nspec) * sizeof(unsigned char));
    {
      int a, b, c, d;
      unsigned char topos, addtopos;
      for (d = 0; d < qblock1.nspec; d++) {
         fprintf(stdout, "%d (max. %d) -> ", d, qblock1.nspec); fflush(stdout);
         fprintf(stdout, "%d (max. %d)\n", qblock1.translat[d], allqblock.nspec); fflush(stdout);
      }
      for (d = 3; d < qblock1.nspec; d++) {
        for (c = 2; c < d; c++)
          for (b = 1; b < c; b++)
            for (a = 0; a < b; a++) {
#ifdef VERBOSE
              fprintf(stdout, "(%d,%d,%d,%d)\t", a, b, c, d);
#endif

              topos = readquartet(qblock1.translat[a], 
                                  qblock1.translat[b], 
                                  qblock1.translat[c], 
                                  qblock1.translat[d], 
                                  allqblock.quartetinfo);
              addtopos = readquartet(a, b, c, d, 
                                  qblock1.quartetinfo);
#ifdef VERBOSE
              fprintf(stdout, "%d | %d\t", topos, addtopos);
#endif
              topos = topos | addtopos;
#ifdef VERBOSE
              fprintf(stdout, "%d\t", topos);
#endif
              writequartet(qblock1.translat[a], 
                                  qblock1.translat[b], 
                                  qblock1.translat[c], 
                                  qblock1.translat[d], 
                                  topos,
                                  allqblock.quartetinfo);
#ifdef VERBOSE
              fprintf(stdout, "-> (%d,%d,%d,%d)\n", qblock1.translat[a], 
                                  qblock1.translat[b], 
                                  qblock1.translat[c], 
                                  qblock1.translat[d] );
#endif
        }

    }

    printf("new sum=%d\n", allqblock.nspec);
    cleanupqblock(&qblock1);
  }

  strcpy(allqblock.fname, "output");
  printbinallquarts(&allqblock);
  cleanupqblock(&allqblock);
}



/******************************************************************************/

void addquartlh(int argc, char *argv[])
{
  qblocktype      qblock1;
  qblocktype      allqblock;
  int             currfile,
                  n1,
                  n2;
  int             sumspec = 0;

  printf(" argc=%d\n", argc);
  initqblock(&allqblock);

  for (currfile=2; currfile < argc; currfile++) {
    initqblock(&qblock1);
    strcpy(qblock1.fname, argv[currfile]);
    /* freshreadallquarts(&qblock1); */
    newreadheader(&qblock1, 1);
    /* printverboseallquarts(stdout, qblock1); */
    
    sumspec += qblock1.nspec;
    cleanupqblock(&qblock1);
  }


  printf("sum=%d\n", sumspec);
  
  allqblock.names = malloc(sumspec * sizeof(char **));
  allqblock.nspec = 0;
  for (currfile=2; currfile < argc; currfile++) {
    initqblock(&qblock1);
    strcpy(qblock1.fname, argv[currfile]);
    newreadheader(&qblock1, 1); 
    qblock1.translat = malloc(qblock1.nspec * sizeof(int));
    for (n1=0; n1<qblock1.nspec; n1++) {
      qblock1.translat[n1] = -1;
      for (n2=0; n2<allqblock.nspec; n2++) {
         if (0 == strcmp(qblock1.names[n1], allqblock.names[n2])) {
           qblock1.translat[n1] = n2;
           fprintf(stdout, "xxx %3d %10s == %3d %10s\n", n1, qblock1.names[n1], n2, allqblock.names[n2]);
         }
      }
      if (qblock1.translat[n1] == -1) {
        allqblock.names[allqblock.nspec] = malloc(11 * sizeof(char));
        strcpy(allqblock.names[allqblock.nspec], qblock1.names[n1]);
        qblock1.translat[n1] = allqblock.nspec;
        allqblock.nspec++;
      }
    }
    cleanupqblock(&qblock1);
  }

  free(allqblock.names);
  allqblock.names = malloc(allqblock.nspec * sizeof(char **));
  allqblock.lhinfo = malloc(3 * numquarts(allqblock.nspec) * sizeof(double));
  allqblock.nspec = 0;

  for (currfile=2; currfile < argc; currfile++) {
    initqblock(&qblock1);
    strcpy(qblock1.fname, argv[currfile]);
    /* newreadheader(&qblock1, 1); */
    newreadallquarts(&qblock1);
    qblock1.translat = malloc(qblock1.nspec * sizeof(int));
    for (n1=0; n1<qblock1.nspec; n1++) {
      qblock1.translat[n1] = -1;
      for (n2=0; n2<allqblock.nspec; n2++) {
         if (0 == strcmp(qblock1.names[n1], allqblock.names[n2])) {
           qblock1.translat[n1] = n2;
           fprintf(stdout, "%3d %10s == %3d %10s\n", n1, qblock1.names[n1], n2, allqblock.names[n2]);
         }
      }
      if (qblock1.translat[n1] == -1) {
        allqblock.names[allqblock.nspec] = malloc(11 * sizeof(char));
        strcpy(allqblock.names[allqblock.nspec], qblock1.names[n1]);
        qblock1.translat[n1] = allqblock.nspec;
        allqblock.nspec++;
      }
    }

/* printf("aaa1\n"); fflush(stdout); */

    {
      int a, b, c, d;
      double d1, d2, d3;
      double s1, s2, s3;

      for (d = 0; d < qblock1.nspec; d++) {
         fprintf(stdout, "%d (max. %d) -> ", d, qblock1.nspec); fflush(stdout);
         fprintf(stdout, "%d (max. %d)\n", qblock1.translat[d], allqblock.nspec); fflush(stdout);
      }
      for (d = 3; d < qblock1.nspec; d++) {
        fprintf(stderr, "%d\n", d);
        for (c = 2; c < d; c++)
          for (b = 1; b < c; b++)
            for (a = 0; a < b; a++) {
#             ifdef VERBOSE
                fprintf(stdout, "(%d,%d,%d,%d)", a, b, c, d);
#             endif

              readquartetlh(qblock1.translat[a], 
                            qblock1.translat[b], 
                            qblock1.translat[c], 
                            qblock1.translat[d], 
                            &s1, &s2, &s3,
                            allqblock.lhinfo);

#             ifdef VERBOSE
                fprintf(stdout, "\t%f,%f,%f <- (%d,%d,%d,%d)\n", 
                                s1, s2, s3, 
                                qblock1.translat[a], 
                                qblock1.translat[b], 
                                qblock1.translat[c], 
                                qblock1.translat[d] );
#             endif

              readquartetlh(a, b, c, d, &d1, &d2, &d3, qblock1.lhinfo);

#             ifdef VERBOSE
                fprintf(stdout, "\t\t%f,%f,%f <- (%d,%d,%d,%d)*\n", 
                                d1, d2, d3, a, b, c, d );
#             endif

              s1 += d1;
              s2 += d2;
              s3 += d3;

              writequartetlh(qblock1.translat[a], 
                             qblock1.translat[b], 
                             qblock1.translat[c], 
                             qblock1.translat[d], 
                             s1, s2, s3,
                             allqblock.lhinfo);

#             ifdef VERBOSE
                fprintf(stdout, "\t\t%f,%f,%f -> (%d,%d,%d,%d)\n", 
                                s1, s2, s3, 
                                qblock1.translat[a], 
                                qblock1.translat[b], 
                                qblock1.translat[c], 
                                qblock1.translat[d] );
#             endif
            } /* for a */
      } /* for d,c,b */
    }

    printf("new sum=%d\n", allqblock.nspec);
    cleanupqblock(&qblock1);
  }

  strcpy(allqblock.fname, "output");
  printbinlh(&allqblock);
  cleanupqblock(&allqblock);
}



/******************************************************************************/

/* unsigned char weightquart(int a, int b, int c, int i) */
unsigned char weightquart(double d1, double d2, double d3)
{
	double temp, temp1, temp2, temp3, onethird;
	unsigned char treebits[3];
	unsigned char discreteweight[3];
	double        qweight[3];
	int           qworder[3];
	double        sqdiff[3];
	int           sqorder[3];
	onethird = 1.0/3.0;
	treebits[0] = (unsigned char) 1;
	treebits[1] = (unsigned char) 2;
	treebits[2] = (unsigned char) 4;

	qweight[0] = d1;
	qweight[1] = d2;
	qweight[2] = d3;

	/* compute_quartlklhds(a,b,c,i,&qweight[0],&qweight[1],&qweight[2], (approxqp ? APPROX : EXACT)); */

	/* sort in descending order */
	sort3doubles(qweight, qworder);

	/* compute Bayesian weights */
	qweight[qworder[1]] = exp(qweight[qworder[1]]-qweight[qworder[0]]);
	qweight[qworder[2]] = exp(qweight[qworder[2]]-qweight[qworder[0]]);
	qweight[qworder[0]] = 1.0;
	temp = qweight[0] + qweight[1] + qweight[2];
	qweight[0] = qweight[0]/temp;
	qweight[1] = qweight[1]/temp;
	qweight[2] = qweight[2]/temp;
							
	/* square deviations */
	temp1 = 1.0 - qweight[qworder[0]];
	sqdiff[0] = temp1 * temp1 +
		    qweight[qworder[1]] * qweight[qworder[1]] +
		    qweight[qworder[2]] * qweight[qworder[2]];
	discreteweight[0] = treebits[qworder[0]];
     
	temp1 = 0.5 - qweight[qworder[0]];
	temp2 = 0.5 - qweight[qworder[1]];
	sqdiff[1] = temp1 * temp1 + temp2 * temp2 +							           
	            qweight[qworder[2]] * qweight[qworder[2]];           
	discreteweight[1] = treebits[qworder[0]] + treebits[qworder[1]];
	temp1 = onethird - qweight[qworder[0]];
	temp2 = onethird - qweight[qworder[1]];
	temp3 = onethird - qweight[qworder[2]];
	sqdiff[2] = temp1 * temp1 + temp2 * temp2 + temp3 * temp3;
	discreteweight[2] = (unsigned char) 7;						           
							
	/* sort in descending order */
	sort3doubles(sqdiff, sqorder);
						
	/* counting completely unresolved quartets */
	if (discreteweight[sqorder[2]] == 7) fprintf(stderr, "unresolved quartet\n");
	else fprintf(stderr, ".");						
	/* determine best discrete weight */
	/* writequartet(a, b, c, i, discreteweight[sqorder[2]]); */
	return(discreteweight[sqorder[2]]);
}




/******************************************************************************/

void printusage(char *argv[])
{
      printf("\nUSAGE: %s -h | -m | -c quartfile [ ... ]\n\n", argv[0]);
      printf("\t-h         - this help text\n");
      printf("\t-c  qf     - check quartfile\n");
      printf("\t-m  qf ... - merge quartfiles\n");
      printf("\t-x  qf qf  - translate binary quartfile to ASCII\n");
      printf("\t-l  lf     - check lhfile\n");
      printf("\t-a  lf ... - add lh files\n");
      printf("\t-b  lf qf  - convert lh file to quartfile\n");
      printf("\t-e  lf lf  - creating empty lh file using header from first lf\n");
      printf("\t-0  lf lf  - creating empty bin lh file using header from first lf\n");
      printf("\t-n  lf lf  - creating empty bin quart file (old format) using header from first lf\n");
      printf("\n");
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/

int main(int argc, char *argv[])
{
   if ((argc < 2) || (0 == strcmp(argv[1], "-h"))) {
      printusage(argv);
      exit(0);
   } else {
      if ((0 == strcmp(argv[1], "-c")) && (argc == 3)) {
         printf("checkquart\n");
         checkquart(argc,argv);
      } else {
         if ((0 == strcmp(argv[1], "-m")) && (argc >= 3)) {
            mergequart(argc,argv);
         } else {
            if ((0 == strcmp(argv[1], "-x")) && (argc == 4)) {
               toasciiquart(argc,argv);
            } else {
               if ((0 == strcmp(argv[1], "-l")) && (argc == 3)) {
                  checklh(argc,argv);
               } else {
                  if ((0 == strcmp(argv[1], "-b")) && (argc == 4)) {
                     lh2quart(argc,argv);
                  } else {
                     if ((0 == strcmp(argv[1], "-a")) && (argc >= 3)) {
                        addquartlh(argc,argv);
                     } else {
                        if ((0 == strcmp(argv[1], "-e")) && (argc == 4)) {
                           emptylh(argc,argv);
                        } else {
                           if ((0 == strcmp(argv[1], "-0")) && (argc == 4)) {
                              emptybinlh(argc,argv);
                           } else {
                              if ((0 == strcmp(argv[1], "-n")) && (argc == 4)) {
                                 emptyoldquart(argc,argv);
                              } else {
                                 printf("\nUnknown command line flag: %s\n", argv[1]);
                                 printf("or unsufficient number of files!!\n");
                                 printusage(argv);
                                 exit(0);
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }
   exit(0);
}
