#include<stdio.h>
#include<stdlib.h>
#include"util.h"
#include<stdlib.h>

#include"newpstep.h"


/******************************************************************************/
/* functions for representing and building puzzling step trees                */
/******************************************************************************/

/* initialize tree with the following starting configuration

                 2
          0  +------- C(=2)
  A(=0) -----+
             +------- B(=1)
                 1
                                */
void newinittree(ONEEDGE **edge, 
              int     **edgeofleaf, 
              int       Maxspc, 
              int       Maxbrnch, 
              int      *nextedge, 
              int      *nextleaf)
{
	int i;

	/* allocate the memory for the whole tree */
	
	/* allocate memory for vector with all the edges of the tree */
	*edge = (ONEEDGE *) calloc(Maxbrnch, sizeof(ONEEDGE) );
	if (*edge == NULL) maerror("edge in inittree");

	/* allocate memory for vector with edge numbers of leaves */ 
	*edgeofleaf = (int *) calloc(Maxspc, sizeof(int) );
	if (*edgeofleaf == NULL) maerror("edgeofleaf in inittree");
	
	/* allocate memory for all the edges the edge map */
	for (i = 0; i < Maxbrnch; i++) {
		(*edge)[i].edgemap = (int *) calloc(Maxbrnch, sizeof(int) );
		if ((*edge)[i].edgemap == NULL) maerror("edgemap in inittree");
	}
		
	/* number all edges */
	for (i = 0; i < Maxbrnch; i++) (*edge)[i].numedge = i;
	
	/* initialize tree */
	
	*nextedge = 3;
	*nextleaf = 3;
	
	/* interconnection */
	(*edge)[0].up        = NULL;
	(*edge)[0].downleft  = &(*edge)[1];
	(*edge)[0].downright = &(*edge)[2];
	(*edge)[1].up        = &(*edge)[0];
	(*edge)[1].downleft  = NULL;
	(*edge)[1].downright = NULL;
	(*edge)[2].up        = &(*edge)[0];
	(*edge)[2].downleft  = NULL;
	(*edge)[2].downright = NULL;
	
	/* edges of leaves */
	(*edgeofleaf)[0] = 0;
	(*edgeofleaf)[1] = 1;
	(*edgeofleaf)[2] = 2;
} /* newinittree */


/************************/

/* add next leaf on the specified edge */
void newaddnextleaf(int      dockedge,
                 ONEEDGE *edge, 
                 int     *edgeofleaf, 
                 int      Maxspc, 
                 int      Maxbrnch, 
                 int     *nextedge, 
                 int     *nextleaf)
{	
	int i;

fprintf(stderr, "docking l%d to n%d\n", *nextleaf, dockedge);

	if (dockedge >= *nextedge) {
		/* Trying to add leaf nextleaf to nonexisting edge dockedge */
		FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR F TO DEVELOPERS\n\n\n");
		exit(1);
	}
	
	if (*nextleaf >= Maxspc) {
		/* Trying to add leaf nextleaf to a tree with Maxspc leaves */
		FPRINTF(STDOUTFILE "\n\n\nHALT: PLEASE REPORT ERROR G TO DEVELOPERS\n\n\n");
		exit(1);
	}
		
	/* necessary change in edgeofleaf if dockedge == edgeofleaf[0] */
	if (edgeofleaf[0] == dockedge) edgeofleaf[0] = *nextedge;
	
	/* adding nextedge to the tree */
	edge[*nextedge].up = edge[dockedge].up;
	edge[*nextedge].downleft = &edge[dockedge];
	edge[*nextedge].downright = &edge[*nextedge+1];
	edge[dockedge].up = &edge[*nextedge];

	if (edge[*nextedge].up != NULL) {
		if ( ((edge[*nextedge].up)->downleft) == &edge[dockedge] ) 
			(edge[*nextedge].up)->downleft  = &edge[*nextedge];
		else   
			(edge[*nextedge].up)->downright = &edge[*nextedge];
	}

	/* adding nextedge + 1 to the tree */
	edge[*nextedge+1].up = &edge[*nextedge];
	edge[*nextedge+1].downleft = NULL;
	edge[*nextedge+1].downright = NULL;
	edgeofleaf[*nextleaf] = *nextedge+1;
	
	(*nextleaf)++;
	*nextedge = *nextedge + 2;
} /* newaddnextleaf */



/************************/

/* free memory (to be called after inittree) */
void newfreetree(ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      Maxspc)
{
	int i;

	for (i = 0; i < 2 * Maxspc - 3; i++) free(edge[i].edgemap);
	free(edge);
	free(edgeofleaf);	
} /* newfreetree */


/************************/

int fputid(FILE *, int);


/************************/

/* writes OTU sitting on edge ed */
void writeOTU(FILE    *outfp, 
              int      ed, 
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *column,
              int     *trueID)
{
	int i;

	/* test whether we are on a leaf */
	if (edge[ed].downright == NULL && edge[ed].downleft == NULL) {
		for (i = 1; i < nextleaf; i++) {
			if (edgeofleaf[i] == ed) { /* i is the leaf of ed */
				*column += fputid(outfp, trueID[i]);
				return;
			}
		}
	}

	/* we are NOT on a leaf */
	fprintf(outfp, "(");
	(*column)++;
	writeOTU(outfp, edge[ed].downleft->numedge,
                 edge, edgeofleaf, nextedge, nextleaf, column, trueID);
	fprintf(outfp, ",");
	(*column)++;
	(*column)++;
	if (*column > 55) {
		*column = 2;
		fprintf(outfp, "\n  ");
	}
	writeOTU(outfp, edge[ed].downright->numedge,	
                 edge, edgeofleaf, nextedge, nextleaf, column, trueID);
	fprintf(outfp, ")");
	(*column)++;
} /* writeOTU */


/************************/

/* write tree */
void writetree(FILE   *outfp,
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *trueID)
{	
	int column;
	column = 1;
	fprintf(outfp, "(");
	column += fputid(outfp, trueID[0]) + 3;
	fprintf(outfp, ",");
	writeOTU(outfp, edge[edgeofleaf[0]].downleft->numedge,
                 edge, edgeofleaf, nextedge, nextleaf, &column, trueID);
	column++;
	column++;
	fprintf(outfp, ",");
	writeOTU(outfp, edge[edgeofleaf[0]].downright->numedge,
                 edge, edgeofleaf, nextedge, nextleaf, &column, trueID);
	fprintf(outfp, ");\n");	
} /* writetree */


/************************/

/* writes OTU sitting on edge ed */
void newwriteOTU(FILE    *outfp, 
              int      ed, 
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *column,
              int     *trueID)
{
	int i;

	/* test whether we are on a leaf */
	if (edge[ed].downright == NULL && edge[ed].downleft == NULL) {
		for (i = 1; i < nextleaf; i++) {
			if (edgeofleaf[i] == ed) { /* i is the leaf of ed */
				*column += fputid(outfp, trueID[i]);
fprintf(outfp, ":%d[n%d,l%d]", edge[ed].edgeinfo, ed, i);
				return;
			}
		}
	}

	/* we are NOT on a leaf */
	fprintf(outfp, "(");
	(*column)++;
	newwriteOTU(outfp, edge[ed].downleft->numedge,
                 edge, edgeofleaf, nextedge, nextleaf, column, trueID);
	fprintf(outfp, ",");
	(*column)++;
	(*column)++;
	if (*column > 55) {
		*column = 2;
		fprintf(outfp, "\n  ");
	}
	newwriteOTU(outfp, edge[ed].downright->numedge,	
                 edge, edgeofleaf, nextedge, nextleaf, column, trueID);
	fprintf(outfp, ")");
fprintf(outfp, ":%d[n%d]", edge[ed].edgeinfo, ed);
	(*column)++;
} /* newwriteOTU */


/************************/

/* write tree */
void newwritetree(FILE   *outfp,
              ONEEDGE *edge, 
              int     *edgeofleaf, 
              int      nextedge, 
              int      nextleaf,
              int     *trueID)
{	
	int column;
	column = 1;
	fprintf(outfp, "(");
	column += fputid(outfp, trueID[0]) + 3;
fprintf(outfp, ":%d[n%d/l%d]", edge[edgeofleaf[0]].edgeinfo, edgeofleaf[0], 0);
	fprintf(outfp, ",");
	newwriteOTU(outfp, edge[edgeofleaf[0]].downleft->numedge,
                 edge, edgeofleaf, nextedge, nextleaf, &column, trueID);
	column++;
	column++;
	fprintf(outfp, ",");
	newwriteOTU(outfp, edge[edgeofleaf[0]].downright->numedge,
                 edge, edgeofleaf, nextedge, nextleaf, &column, trueID);
	fprintf(outfp, ");\n");	
} /* newwritetree */



/************************/

/* clear all edgeinfos */
void newresetedgeinfo(ONEEDGE *edge, 
                   int      nextedge)
{
	int i;
	
	for (i = 0; i < nextedge; i++)
		edge[i].edgeinfo = 0;
} /* newresetedgeinfo */


/************************/

/* recursive function to get bipartitions */
int computeinfos(ONEEDGE  *curredge,
                 ONEEDGE  *edge, 
                 int      *edgeofleaf, 
                 ulimatrix mat,
                 ulivector vec,
                 int      *trueID)
{       
  int n, m, lf;
  int rightnum = 0; 
  int leftnum  = 0; 
  ONEEDGE *rightedge;
  ONEEDGE *leftedge;
        
  if ((curredge->downright == NULL)||
      (curredge->downleft  == NULL)) { /* if curredge is leaf */


       for (lf = 1; edgeofleaf[lf] != curredge->numedge; lf++) {}
           /* lf is the leaf of curredge */

       curredge->edgemap[0] = lf;
       curredge->edgeinfo = vec[lf];
       curredge->edgesum  = vec[lf];
       curredge->edgemat  = 0;
       return(1);

   } else {
     rightnum = computeinfos(curredge->downright, edge, edgeofleaf, 
                             mat, vec, trueID);
     rightedge = curredge->downright;
     curredge->edgesum  = rightedge->edgesum;
     curredge->edgemat  = rightedge->edgemat;
     for (n=0; n<rightnum; n++) {
        curredge->edgemap[n] = rightedge->edgemap[n];
     }

     leftnum  = computeinfos(curredge->downleft, edge, edgeofleaf, 
                             mat, vec, trueID);
     leftedge = curredge->downleft;
     curredge->edgesum += leftedge->edgesum;
     curredge->edgemat += leftedge->edgemat;
     for (m=0; m<leftnum; m++) {
       for (n=0; n<rightnum; n++) {
           curredge->edgemat += 
               2*( mat[leftedge->edgemap[m]] [rightedge->edgemap[n]]);
        }
        curredge->edgemap[rightnum+m] = leftedge->edgemap[m];
     }

     curredge->edgeinfo = curredge->edgesum - curredge->edgemat;

     return(rightnum + leftnum);
   }
   
} /* computeinfos */


/************************/

void newincrementedgeinfo(ONEEDGE  *edge, 
                          int      *edgeofleaf, 
                          ulimatrix mat,
                          ulivector vec,
                          int      *trueID)
{
   computeinfos(&(edge[edgeofleaf[0]]), edge, edgeofleaf, mat, vec, trueID);
   if (edge[edgeofleaf[0]].edgeinfo != vec[0])
      fprintf(stderr, "ERROR: root penalty (%lu) != recursive penalty (%lu)\n",
              vec[0], edge[0].edgeinfo);

} /* newincrementedgeinfo */


/************************/

/* checks which edge has the lowest edgeinfo
   if there are several edges with the same lowest edgeinfo,
   one of them will be selected randomly */
void newminimumedgeinfo(ONEEDGE *edge, 
                     int     *edgeofleaf, 
                     int      nextedge, 
                     int      nextleaf,
                     int     *minedge, 
                     uli     *mininfo, 
                     int     *howmany) 
{
	int i, k, randomnum;

	*howmany = 1;
	*minedge = 0;
	*mininfo = edge[0].edgeinfo;
	for (i = 1; i < nextedge; i++)
		if (edge[i].edgeinfo <= *mininfo) {
			if (edge[i].edgeinfo == *mininfo) {
				(*howmany)++;
			} else {
				*minedge = i;
				*mininfo = edge[i].edgeinfo;
				*howmany = 1;
			}
		}
	
	if (*howmany > 1) { /* draw random edge */
		randomnum = randominteger(*howmany) + 1; /* 1 to howmany */
		i = -1;
		for (k = 0; k < randomnum; k++) {
			do {
				i++;
			} while (edge[i].edgeinfo != *mininfo);
			*minedge = i;
		}
	}
} /* newminimumedgeinfo */


/************************/
